;===============================================================
;       EF12, Version 2.1, 18apr00cht
;               move to 8000H replacing WHYP.
;               copy interrupt vectors from WHYPFLSH.S19
;               to EF12.S19 before flashing
;               add TICKS1 and DELAY1 for motor stepping
;
;       EF12, 02/18/00, C. H. Ting
;       Adapt 86eForth v2.02 to 68HC12.
;               Use WHYP to seed EF12.ASM
;               Use AS12 native 68HC12 assembler:
;               as12 ef12.asm >ef12.lst
;       EF12A, add ADC code, 02mar00cht
;       EF12B, 01mar00cht
;               stack to $78, return stack to $f8.
;               add all port definitions
;               add PWM registers
;               add SPI registers and code
;       EF12C, 12mar00cht
;               add MAX5250 D/A converter
;       EF12D, 15mar00cht
;               add all the Lexel interface words
;       EF12E, 18apr00cht, save for reference
;
;       Copyright (c) 2000
;       Dr. C. H. Ting
;       156 14th Avenue
;       San Mateo, CA 94402
;       (650) 571-7639
;
;===============================================================

*********************************************************
REGBASE EQU	$0000	   ;register base
RAMBASE EQU	$0800	   ;ram base
EESTART EQU	$0D00	   ;start address of EEPROM
EESTOP  EQU	$0FFF	   ;stop address of EEPROM
WHYPBASE EQU    $8000      ;start of WHYP kernel
***********************************************************
START   EQU	COLD
***********************************************************
SC0BDL  EQU	REGBASE+$C1	;baud rate control
SC0CR1  EQU	REGBASE+$C2	;SCI control reg 1
SC0CR2  EQU	REGBASE+$C3	;SCI control reg 2
SC0SR   EQU	REGBASE+$C4	;SCI status reg
SC0DR   EQU	REGBASE+$C7	;SCI data reg

COPCTL  EQU	REGBASE+$16	;COP control reg
EEPROT  EQU	REGBASE+$F1	;EEPROM Block Protect reg
EEPROG  EQU	REGBASE+$F3	;EEPROM programming control reg
RDRF	EQU	$20		;SCSR mask
STACK   EQU	RAMBASE+$F8	;system (return) stack 
DATSTK  EQU	RAMBASE+$78	;data stack 
******  WHYP System Variables  ******
SP0	EQU	RAMBASE	 ;initial data stack pointer
RP0	EQU	RAMBASE+$02	;initial return stack pointer
*LAST	EQU	RAMBASE+$04	;address called by CREATE
***********************************************

        ORG     WHYPBASE
;; Version control

VER     EQU     2         ;major release version
EXT     EQU     1         ;minor extension

;; Constants

TRUEE   EQU     -1      ;true flag

COMPO   EQU     $40     ;lexicon compile only bit
IMEDD   EQU     $80     ;lexicon immediate bit
MASKK   EQU     $1F7F  ;lexicon bit mask

CELLL   EQU     2       ;size of a cell
BASEE   EQU     10      ;default radix

BKSPP   EQU     8       ;back space
LF      EQU     10      ;line feed
CRR     EQU     13      ;carriage return
ERR     EQU     27      ;error escape
TIC     EQU     39      ;tick

CALLL   EQU     $16     ;JSR opcodes

;; Memory allocation

UPP     EQU     RAMBASE+6
SPP     EQU     RAMBASE+$78
RPP     EQU     RAMBASE+$F8
TIBB    EQU     RAMBASE+$80
CTOP    EQU     RAMBASE+$100

;; Main entry points and COLD start data

ORIG:   
MAIN    
        LDS     #STACK  ;initialize return stack
        STS     RP0
        LDX     #DATSTK ;initialize data stack
        STX     SP0
        CLR     EEPROT  ;enable EEPROM writes
        CLR     COPCTL  ;disable COP watchdog
        CLR     SC0CR1  ;8 bit
        LDD     #$340C
        STAA    SC0BDL  ;9600 baud
        STAB    SC0CR2  ;enable tx & rx
        JMP     START   ;default=MN1

; COLD start initiates these variables.

UZERO:
        DW      BASEE   ;BASE
        DW      0       ;tmp
        DW      0       ;>IN
        DW      0       ;#TIB
        DW      TIBB    ;TIB
        DW      INTER   ;'EVAL
        DW      0       ;HLD
        DW       LASTN   ;CONTEXT pointer
        DW       CTOP   ;CP in RAM
        DW      LASTN   ;LAST
ULAST:  DW      0

;; Device dependent I/O
;       All channeled to DOS 21H services

;       ?RX     ( -- c T | F )
;         Return input byte and true, or false.
        LOC
        DW      0
LINK`
        DB      4
        FCC     '?KEY'
QKEY
        LDAA    SC0SR   ;check status
        ANDA    #RDRF   ;check rdrf
        BEQ     INCH    ;wait for char
        LDAB    SC0DR   ;get char in A
        CLRA
        STD     2,-X
        LDD     #-1
        STD     2,-X
        RTS
INCH    CLRB
        STD     2,-X
        RTS

;       TX!     ( c -- )
;         Send character c to  output device.
        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'EMIT'
EMIT
        LDD     2,X+
        TBA
OUTPUT  
        TST     SC0SR
        BPL     OUTPUT  ;loop until tdre
        STAA    SC0DR   ;send A
        RTS

;; The kernel

;       doLIT   ( -- w )
;       Push an inline literal.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+5
        FCC     'doLit'
DOLIT
        PULY
        LDD     0,Y
        STD     2,-X
        INY
        INY
        PSHY
        RTS

;       next    ( -- )
;         Code for  single index loop.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+4
        FCC     'next'
DONXT
        PULY
        PULD
        SUBD    #1
        BCC     NEX1
        JMP     2,Y
NEX1    PSHD
        JMP     [0,Y]

;       ?branch ( f -- )
;       Branch if flag is zero.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+7
        FCC     '?branch'
QBRAN
        PULY
        LDD     2,X+
        BEQ     QBR1
        JMP     2,Y
QBR1    JMP     [0,Y]
        
;       branch  ( -- )
;       Branch to an inline address.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+6
        FCC     'branch'
BRAN
        PULY
        JMP     [0,Y]

;       EXECUTE ( ca -- )
;         Execute  word at ca.

        DW      LINK`
        LOC
LINK`
        DB       7
        FCC     'EXECUTE'
EXECU
        LDY     2,X+
        JMP     0,Y

;       EXIT    ( -- )
;       Terminate a colon definition.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'EXIT'
EXIT
        PULY
        RTS

;       !       ( w a -- )
;         Pop  data stack to memory.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '!'
STORE
        LDY     2,X+    ;Y=a
        MOVW    2,X+,0,Y ;store w at a
        RTS     

;       @       ( a -- w )
;         Push memory location to stack.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '@'
AT
        LDY     0,X     ;Y = a
        MOVW    0,Y,0,X ;w = @Y
        RTS     

;       C!      ( c b -- )
;         Pop  data stack to byte memory.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'C!'
CSTOR
        LDY     2,X+    ;Y=b
        LDD     2,X+    ;D = c
        STAB    0,Y     ;store c at b
        RTS     

;       C@      ( b -- c )
;         Push byte in memory to  stack.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'C@'
CAT
        LDY     0,X     ;Y=b
        CLRA    
        LDAB    0,Y
        STD     0,X
        RTS     

;       RP@     ( -- a )
;         Push current RP to data stack.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'rp@'
RPAT
        PULD    ;save return addr
        TFR     SP,Y    ;SP -> Y
        STY     2,-X    ;push Y to stack
        PSHD    ;restore return addr
        RTS     

;       RP!     ( a -- )
;         Set  return stack pointer.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+3
        FCC     'rp!'
RPSTO
        PULD    ;save return addr
        LDY     2,X+    ;Y=a
        TFR     Y,SP    ;Y -> SP, SP = a
        PSHD    ;restore return addr
        RTS     

;       R>      ( -- w )
;         Pop return stack to data stack.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+2
        FCC     'R>'
RFROM
        PULY    ;save return addr
        PULD    ;D = ret stack val, w
        STD     2,-X    ;push w on stack
        PSHY    ;restore return addr
        RTS     

;       R@      ( -- w )
;         Copy top of return stack to stack.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'R@'
RAT
        TSY     ;SP+1 -> Y
        LDD     2,Y     ;D = ret w
        STD     2,-X    ;push w on stack
        RTS     

;       >R      ( w -- )
;         Push data stack to return stack.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+2
        FCC     '>R'
TOR
        PULY    ;save return addr
        LDD     2,X+    ;D = w
        PSHD    ;push D on ret stack
        PSHY    ;restore return addr
        RTS     

;       SP@     ( -- a )
;         Push current stack pointer.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'sp@'
SPAT
        PSHX
        PULY    ;Y = X = a
        STY     2,-X    ;push on stack
        RTS     

;       SP!     ( a -- )
;         Set  data stack pointer.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'sp!'
SPSTO
        LDX     0,X     ;X = a
        RTS     

;       DROP    ( w -- )
;       Discard top stack item.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DROP'
DROP
        INX     
        INX     
        RTS     

;       DUP     ( w -- w w )
;         Duplicate  top stack item.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'DUP'
DUPP
        MOVW 0,X,2,-X   
        RTS     

;       SWAP    ( w1 w2 -- w2 w1 )
;       Exchange top two stack items.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'SWAP'
SWAP
        LDD 0,X 
        MOVW 2,X,0,X    
        STD 2,X 
        RTS     

;       OVER    ( w1 w2 -- w1 w2 w1 )
;       Copy second stack item to top.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'OVER'
OVER
        MOVW 2,X,2,-X   
        RTS     

;       0<      ( n -- t )
;       Return true if n is negative.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '0<'
ZLESS
        LDY     #$FFFF  ;true
        LDD     0,X     ;if positive
        BMI     ZL1
        LDY     #0000   ;false
ZL1     STY     0,X
        RTS     

;       AND     ( w w -- w )
;       Bitwise AND.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'AND'
ANDD
        LDD     2,X+    ;D=w, pop stack to D
        ANDA    0,X
        STAA    0,X
        ANDB    1,X
        STAB    1,X
        RTS

;       OR      ( w w -- w )
;       Bitwise inclusive OR.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'OR'
ORR
        LDD     2,X+    ;D=w
        ORAA    0,X
        STAA    0,X
        ORAB    1,X
        STAB    1,X
        RTS

;       XOR     ( w w -- w )
;       Bitwise exclusive OR.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'XOR'
XORR
        LDD     2,X+    ;D=w
        EORA    0,X
        STAA    0,X
        EORB    1,X
        STAB    1,X
        RTS

;       UM+     ( u u -- udsum )
;         Add two unsigned single
;         and return a double sum.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'UM+'
UPLUS
        LDY     #1
        LDD     2,X
        ADDD    0,X
        STD     2,X
        BCS     UPL1
        LDY     #0
UPL1    STY     0,X
        RTS

;; System and user variables

;       doVAR   ( -- a )
;         Code for VARIABLE and CREATE.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+5
        FCC     'doVar'
DOVAR
        PULY    ;get return addr (pfa)
        STY     2,-X    ;push on stack
        RTS     ;go to RTS of EXEC

;       UP      ( -- a )
;         Pointer to  user area.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'up'
UP
        LDD     #(RAMBASE+6)
        STD     2,-X
        RTS

;       BASE    ( -- a )
;         Radix base for numeric I/O.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'BASE'
BASE
        LDD     #(RAMBASE+6)
        STD     2,-X
        RTS

;       tmp     ( -- a )
;         A temporary storage.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+3
        FCC     'tmp'
TEMP
        LDD     #(RAMBASE+8)
        STD     2,-X
        RTS

;       >IN     ( -- a )
;         Hold parsing pointer.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     '>IN'
INN
        LDD     #(RAMBASE+10)
        STD     2,-X
        RTS

;       #TIB    ( -- a )
;         Count in terminal input buffer.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     '#TIB'
NTIB
        LDD     #(RAMBASE+12)
        STD     2,-X
        RTS

;       'EVAL   ( -- a )
;       Execution vector of EVAL.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     "'eval"
TEVAL
        LDD     #(RAMBASE+16)
        STD     2,-X
        RTS


;       HLD     ( -- a )
;         Hold a pointer of output string.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'hld'
HLD
        LDD     #(RAMBASE+18)
        STD     2,-X
        RTS

;       CONTEXT ( -- a )
;         Start vocabulary search.

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'CONTEXT'
CNTXT
        LDD     #(RAMBASE+20)
        STD     2,-X
        RTS

;       CP      ( -- a )
;         Point to top of dictionary.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'cp'
CP
        LDD     #(RAMBASE+22)
        STD     2,-X
        RTS

;       LAST    ( -- a )
;         Point to last name in dictionary.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'last'
LAST
        LDD     #(RAMBASE+24)
        STD     2,-X
        RTS

;; Common functions

;       ?DUP    ( w -- w w | 0 )
;       Dup tos if its is not zero.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     '?DUP'
QDUP
        LDD     0,X
        BEQ     QDUP1
        STD     2,-X
QDUP1:  RTS

;       ROT     ( w1 w2 w3 -- w2 w3 w1 )
;       Rot 3rd item to top.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'ROT'
ROT
        LDD     0,X     ;D = n3
        MOVW    4,X,0,X ;n1 n2 n1
        MOVW    2,X,4,X ;n2 n2 n1
        STD     2,X     ;n2 n3 n1
        RTS

;       2DROP   ( w w -- )
;       Discard two items on stack.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     '2DROP'
DDROP
        INX
        INX
        INX
        INX
        RTS

;       2DUP    ( w1 w2 -- w1 w2 w1 w2 )
;       Duplicate top two items.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     '2DUP'
DDUP
        MOVW    2,X,2,-X
        MOVW    2,X,2,-X
        RTS

;       +       ( w w -- sum )
;       Add top two items.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '+'
PLUS
        LDD     2,X+
        ADDD    0,X
        STD     0,X
        RTS

;       NOT     ( w -- w )
;       One's complement of tos.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'NOT'
INVER
        COM     0,X
        COM     1,X
        RTS

;       NEGATE  ( n -- -n )
;       Two's complement of tos.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'NEGATE'
NEGAT
        NEG     1,X     ;negate low byte
        BEQ     NG1     ;if not zero
        COM     0,X     ;complement hi byte
        RTS
NG1     NEG     0,X     ;else negate hi byte
        RTS

;       DNEGATE ( d -- -d )
;       Two's complement of top double.

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'DNEGATE'
DNEGA
        NEG     3,X     ;dnegate
        BNE     DN1
        NEG     2,X
        BNE     DN2
        NEG     1,X
        BNE     DN3
        NEG     0,X
        RTS
DN1     COM     2,X
DN2     COM     1,X
DN3     COM     0,X
        RTS

;       -       ( n1 n2 -- n1-n2 )
;       Subtraction.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '-'
SUBB
        LDD     2,X
        SUBD    2,X+
        STD     0,X
        RTS

;       ABS     ( n -- n )
;         Return  absolute value of n.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'ABS'
ABSS
        TST     0,X     ;if negative
        BPL     AB2     ;negate:
        NEG     1,X     ;negate low byte
        BEQ     AB1     ;if not zero
        COM     0,X     ;complement hi byte
        RTS
AB1     NEG     0,X     ;else negate hi byte
AB2     RTS

;       =       ( w w -- t )
;       Return true if top two are equal.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '='
EQUAL
        LDY     #$FFFF  ;true
        LDD     2,X+    ;D = n2
        CPD     0,X     ;if n2 <> n1
        BEQ     EQ1
        LDY     #0      ;set false
EQ1     STY     0,X     ;set flag
        RTS     

;       U<      ( u u -- t )
;       Unsigned compare of top two items.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'U<'
ULESS
        LDY     #$FFFF  ;true
        LDD     2,X+    ;D = u2
        CPD     0,X     ;if u2 <= u1
        BHI     ULT1
        LDY     #0      ;set false
ULT1    STY     0,X     ;set flag
        RTS     

;       <       ( n1 n2 -- t )
;       Signed compare of top two items.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '<'
LESS
        LDY     #$FFFF  ;true
        LDD     2,X+    ;D = n2
        CPD     0,X     ;if n2 <= n1
        BGT     LT1
        LDY     #0      ;set false
LT1     STY     0,X     ;set flag
        RTS     

;       MAX     ( n n -- n )
;         Return greater of two top items.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'MAX'
MAX
        LDY     2,X+    ;Y = n2
        CPY     0,X     ;if n2 > n1
        BLE     MX1
        STY     0,X     ;leave n2 on stack
MX1     RTS     ;else leave n1

;       MIN     ( n n -- n )
;         Return smaller of top two items.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'MIN'
MIN
        LDY     2,X+    ;Y = n2
        CPY     0,X     ;if n2 < n1
        BGE     MIN1
        STY     0,X     ;leave n2 on stack
MIN1    RTS     ;else leave n1

;       WITHIN  ( u ul uh -- t )
;         Return true if u is within
;         range of ul and uh. ( ul <= u < uh )

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'WITHIN'
WITHI
        JSR     OVER
        JSR     SUBB
        JSR     TOR
        JSR     SUBB
        JSR     RFROM
        JMP     ULESS

;; Divide

;       UM/MOD  ( udl udh un -- ur uq )
;       Unsigned divide of a double by a
;         single. Return mod and quotient.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'UM/MOD'
UMMOD
        PSHX
        LDY     2,X
        LDD     4,X
        LDX     0,X
        EDIV
        BCS     UM1             ;if div by 0
        BVC     UM2             ; or overflow
UM1     LDD     #$FFFF    ; rem = $FFFF
        LDY     #$FFFF    ; quot = $FFFF
UM2     PULX
        LEAX    2,X
        STD     2,X
        STY     0,X
        RTS

;       M/MOD   ( d n -- r q )
;       Signed floored divide of double by
;         single. Return mod and quotient.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'M/MOD'
MSMOD
        PSHX
        LDY     2,X
        LDD     4,X
        LDX     0,X
        EDIVS
        BCS     MM1     ;if div by 0
        BVC     MM2     ; or overflow
MM1     LDD     #$FFFF  ; rem = $FFFF
        LDY     #$FFFF  ; quot = $FFFF
MM2     PULX
        LEAX    2,X
        STD     2,X
        STY     0,X
        RTS

;       /MOD    ( n n -- r q )
;       Signed divide. Return mod and quotient.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     '/MOD'
SLMOD
        PSHX
        LDD     2,X
        LDX     0,X
        IDIVS
        TFR     X,Y
        BCS     SM1     ;if div by 0
        BVC     SM2     ; or overflow
SM1     LDD     #$FFFF  ; rem = $FFFF
        LDY     #$FFFF  ; quot = $FFFF
SM2     PULX
        STD     2,X
        STY     0,X
        RTS

;       MOD     ( n n -- r )
;       Signed divide. Return mod only.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'MOD'
MODD
        PSHX
        LDD     2,X
        LDX     0,X
        IDIVS
        TFR     X,Y
        BCS     MD1             ;if div by 0
        BVC     MD2             ; or overflow
MD1     LDD     #$FFFF    ; rem = $FFFF
MD2     PULX
        LEAX    2,X
        STD     0,X
        RTS

;       /       ( n n -- q )
;       Signed divide. Return quotient only.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '/'
SLASH
        PSHX
        LDD     2,X
        LDX     0,X
        IDIVS
        TFR     X,Y
        BCS     SL1     ;if div by 0
        BVC     SL2     ; or overflow
SL1     LDD     #$FFFF  ; rem = $FFFF
        LDY     #$FFFF  ; quot = $FFFF
SL2     PULX
        LEAX    2,X
        STY     0,X
        RTS

;; Multiply

;       UM*     ( u u -- ud )
;       Unsigned multiply. Return double product.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'UM*'
UMSTA
        LDD     2,X+
        LDY     2,X+
        EMUL
        STD     2,-X
        STY     2,-X
        RTS     

;       *       ( n n -- n )
;       Signed multiply. Return single product.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '*'
STAR
        LDD     2,X+
        LDY     2,X+
        EMUL
        STD     2,-X
        RTS

;       M*      ( n n -- d )
;       Signed multiply. Return double product.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'M*'
MSTAR      
        LDD     2,X+
        LDY     2,X+
        EMULS
        STD     2,-X
        STY     2,-X
        RTS

;       */MOD   ( n1 n2 n3 -- r q )
;       Multiply n1 and n2, then divide
;         by n3. Return mod and quotient.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     '*/MOD'
SSMOD
        JSR     TOR
        JSR     MSTAR
        JSR     RFROM
        JMP     MSMOD

;       */      ( n1 n2 n3 -- q )
;         Multiply n1 by n2, then divide
;         by n3. Return quotient only.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '*/'
STASL
        PSHX
        LDD     4,X
        LDY     2,X
        EMULS
        LDX     0,X
        EDIVS
        BCS     SS1     ;if div by 0
        BVC     SS2     ; or overflow
SS1     LDY     #$FFFF  ; quot = $FFFF
SS2     PULX
        LEAX    4,X
        STY     0,X
        RTS

;; Miscellaneous

;       CELL+   ( a -- a )
;       Add cell size in byte to address.

        DW      LINK`
        LOC
LINK`
        DB       2
        FCC     '2+'
CELLP
        LDD     0,X
        ADDD    #2
        STD     0,X
        RTS

;       CELL-   ( a -- a )
;         Subtract 2 from address.

        DW      LINK`
        LOC
LINK`
        DB       2
        FCC     '2-'
CELLM
        LDD     0,X
        SUBD    #2
        STD     0,X
        RTS

;       CELLS   ( n -- n )
;         Multiply tos by 2.

        DW      LINK`
        LOC
LINK`
        DB       2
        FCC     '2*'
CELLS
        ASL     1,X     ;arith shift left
        ROL     0,X
        RTS

;       1+      ( a -- a )
;       Add cell size in byte to address.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '1+'
ONEP
        LDD     0,X
        ADDD    #1
        STD     0,X
        RTS

;       1-      ( a -- a )
;         Subtract 2 from address.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '1-'
ONEM
        LDD     0,X
        SUBD    #1
        STD     0,X
        RTS

;       2/      ( n -- n )
;         Multiply tos by 2.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '2/'
TWOSL
        ASR     0,X     ;arith shift right
        ROR     1,X
        RTS

;       BL      ( -- 32 )
;         Return 32,  blank character.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'BL'
BLANK
        LDD     #32
        STD     2,-X
        RTS

;         0     ( -- 0)
;         Return 0.

        DW      LINK`
        LOC
LINK`
        DB       1
        FCC     '0'
ZERO
        LDD     #0
        STD     2,-X
        RTS

;         1     ( -- 1)
;         Return 1.

        DW      LINK`
        LOC
LINK`
        DB       1
        FCC     '1'
ONE
        LDD     #1
        STD     2,-X
        RTS

;         -1    ( -- -1)
;         Return 32,  blank character.

        DW      LINK`
        LOC
LINK`
        DB       2
        FCC     '-1'
MONE
        LDD     #-1
        STD     2,-X
        RTS

;       >CHAR   ( c -- c )
;       Filter non-printing characters.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     '>CHAR'
TCHAR
        JSR     DOLIT
        DW       $7F
        JSR     ANDD
        JSR     DUPP    ;mask msb
        JSR     DOLIT
        DW      127
        JSR     BLANK
        JSR     WITHI   ;check for printable
        JSR     QBRAN
        DW      TCHA1
        JSR     DROP
        JSR     DOLIT
        DW      '_'     ;replace non-printables
TCHA1:  RTS

;       DEPTH   ( -- n )
;         Return  depth of  data stack.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'DEPTH'
DEPTH
        STX     2,-X    ;save data stack ptr
        LDD     SP0
        SUBD    0,X     ;#bytes = SP0 - X
        ASRA
        RORB    ;D = #stack items
        STD     0,X     ; if neg, underflow
        RTS

;       PICK    ( ... +n -- ... w )
;         Copy  nth stack item to tos.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'PICK'
PICK
        LDD     2,X+    ;D = n1
        NOP
        LSLD
        MOVW    D,X,2,-X
        RTS

;; Memory access

;       +!      ( n a -- )
;         Add n to  contents at address a.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '+!'
PSTOR
        LDY     2,X+    ;Y = addr
        LDD     0,Y     ;D = @addr
        ADDD    2,X+    ;add n to @addr
        STD     0,Y
        RTS

;       2!      ( d a -- )
;         Store  double integer to address a.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '2!'
DSTOR
        LDY     2,X+    ;Y = a
        MOVW    2,X+,2,Y+       ;store dH
        MOVW    2,X+,0,Y        ;store dL
        RTS

;       2@      ( a -- d )
;       Fetch double integer from address a.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '2@'
DAT
        LDY     2,X+    ;Y = a
        MOVW    2,Y,2,-X        ;push dL on data stack
        MOVW    0,Y,2,-X        ;push dH on data stack
        RTS

;       COUNT   ( b -- b +n )
;       Return count byte of a string
;         and add 1 to byte address.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'COUNT'
COUNT
        JSR     DUPP
        JSR     ONEP
        JSR     SWAP
        JMP     CAT

;       HERE    ( -- a )
;         Return  top of  code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'HERE'
HERE
        JSR     CP
        JMP     AT

;       PAD     ( -- a )
;         Return address of text buffer
;         above  code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'PAD'
PAD
        JSR     HERE
        JSR     DOLIT
        DW      80
        JMP     PLUS

;       TIB     ( -- a )
;         Return address of terminal input buffer.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'TIB'
TIB
        JSR     NTIB
        JSR     CELLP
        JMP     AT

;       @EXECUTE        ( a -- )
;       Execute vector stored in address a.

        DW      LINK`
        LOC
LINK`
        DB      8
        FCC     '@EXECUTE'
ATEXE
        JSR     AT
        JSR     QDUP    ;?address or zero
        JSR     QBRAN
        DW      EXE1
        JSR     EXECU   ;execute if non-zero
EXE1:   RTS     ;do nothing if zero

;       CMOVE   ( b1 b2 u -- )
;       Copy u bytes from b1 to b2.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'CMOVE'
CMOVE
        LDD     2,X+    ; D = count
        LDY     2,X+    ; Y = to address
        PSHX    ; save X
        LDX     0,X     ; X = from address
CMV1
        SUBD    #1      ; move cnt bytes
        BCS     CMV2
        MOVB    1,X+,1,Y+         ; move bytes from X to Y
        BRA     CMV1
CMV2    PULX    ; restore X
        INX     ; pop data stack
        INX
        RTS

;       FILL    ( b u c -- )
;       Fill u bytes of character c
;         to area beginning at b.

        DW       LINK`
        LOC
LINK`
        DB       4
        FCC     'FILL'
FILL
        JSR     SWAP
        JSR     TOR
        JSR     SWAP
        JSR     BRAN
        DW      FILL2
FILL1:  JSR     DDUP
        JSR     CSTOR
        JSR     ONEP
FILL2:  JSR     DONXT
        DW      FILL1
        JMP     DDROP

;       ERASE   ( b u -- )
;       Erase u bytes beginning at b.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'ERASE'
ERASE
        JSR     ZERO
        JMP     FILL

;       PACK$   ( b u a -- a )
;       Build a counted string with
;         u characters from b. Null fill.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PACK$'
PACKS
        JSR     DUPP
        JSR     TOR     ;strings only on cell boundary
        JSR     DDUP
        JSR     CSTOR
        JSR     ONEP ;save count
        JSR     SWAP
        JSR     CMOVE
        JSR     RFROM
        RTS

;         MASK  8-bit byte
;         INPUT: B = bit no. 
;         OUTPUT; B = mask = 2^bit#
MASK
        PSHY                    ;save Y
        BSR     MSK1
        DB       1
        DB       2
        DB       4
        DB       8
        DB       16
        DB       32
        DB       64
        DB       128
MSK1    PULY                    ;addr of DB  1
        ABY
        LDAB    0,Y             ;get mask
        PULY                    ;restore Y
        RTS

;         HI     ( b# addr -- )
;         set bit number b# of byte at address addr to 1
        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'HI'
HIGH
        LDY     2,X+            ;Y = addr
        LDD     2,X+            ;B = b#
        BSR     MASK            ;B = mask
        STAB    1,-X            ;push B on data stack
        LDAA    0,Y             ;A = @Y
        ORAA    1,X+            ;OR with mask and pop mask
        STAA    0,Y             ;store back at addr
        RTS

;         LO     ( b# addr -- )
;         clear bit number b# of byte at address addr to 0
        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'LO'
LOW
        LDY     2,X+            ;Y = addr
        LDD     2,X+            ;B = b#
        BSR     MASK            ;B = mask 
        COMB                    ;complement mask
        STAB    1,-X            ;push B on data stack
        LDAA    0,Y             ;A = @Y
        ANDA    1,X+            ;AND with mask and pop mask
        STAA    0,Y             ;store back at addr
        RTS

;         ?HI    ( b# addr -- f )
;         leave a true flag if bit number b# of byte at address addr is high
        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     '?HI'
QHI     
        LDY     2,X+            ;Y = addr
        LDD     2,X+            ;B = b#
        BSR     MASK            ;b = mask 
        STAB    1,-X            ;push B on data stack
        LDAA    0,Y             ;A = @Y
        ANDA    1,X+            ;AND with mask and pop mask 
        BEQ     QH1             ;if not zero
        LDD     #$FFFF    ; leave true flag
        BRA     QH2
QH1     LDD     #$0000    ;else leave false flag
QH2     STD     2,-X            ;push flag on stack
        RTS

;; Numeric output, single precision

;       DIGIT   ( u -- c )
;       Convert digit u to a character.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'DIGIT'
DIGIT
        LDA     1,X
        CMPA    #9
        BLS     HA1
        ADDA    #$37
        STA     1,X
        RTS
HA1     ADDA    #$30
        STA     1,X
        RTS

;       EXTRACT ( n base -- n c )
;         Extract least significant digit from n.

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'EXTRACT'
EXTRC
        JSR     ZERO
        JSR     SWAP
        JSR     UMMOD
        JSR     SWAP
        JMP     DIGIT

;       <#      ( -- )
;         Initiate  numeric output process.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '<#'
BDIGS
        JSR     PAD
        JSR     HLD
        JMP     STORE

;       HOLD    ( c -- )
;         Insert a character into output string.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'HOLD'
HOLD
        JSR     HLD
        JSR     AT
        JSR     ONEM
        JSR     DUPP
        JSR     HLD
        JSR     STORE
        JMP     CSTOR

;       #       ( u -- u )
;         Extract one digit from u and
;         append digit to output string.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '#'
DIG
        JSR     BASE
        JSR     AT
        JSR     EXTRC
        JMP     HOLD

;       #S      ( u -- 0 )
;         Convert u until all digits
;         are added to output string.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '#S'
DIGS
DIGS1:  JSR     DIG
        JSR     DUPP
        JSR     QBRAN
        DW      DIGS2
        BRA     DIGS1
DIGS2:  RTS

;       SIGN    ( n -- )
;         Add a minus sign to
;         numeric output string.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'SIGN'
SIGN
        JSR     ZLESS
        JSR     QBRAN
        DW      SIGN1
        JSR     DOLIT
        DW      '-'
        JMP     HOLD
SIGN1:  RTS

;       #>      ( w -- b u )
;         Prepare output string.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '#>'
EDIGS
        JSR     DROP
        JSR     HLD
        JSR     AT
        JSR     PAD
        JSR     OVER
        JMP     SUBB

;       str     ( w -- b u )
;       Convert a signed integer
;         to a numeric string.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'str'
STR
        JSR     DUPP
        JSR     TOR
        JSR     ABSS
        JSR     BDIGS
        JSR     DIGS
        JSR     RFROM
        JSR     SIGN
        JMP     EDIGS

;       HEX     ( -- )
;       Use radix 16 as base for
;         numeric conversions.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'HEX'
HEX
        JSR     DOLIT
        DW      16
        JSR     BASE
        JMP     STORE

;       DECIMAL ( -- )
;       Use radix 10 as base
;         for numeric conversions.

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'DECIMAL'
DECIM
        JSR     DOLIT
        DW      10
        JSR     BASE
        JMP     STORE

;; Numeric input, single precision

;       DIGIT?  ( c base -- u t )
;       Convert a character to its numeric
;         value. A flag indicates success.

        DW      LINK`
        LOC
LINK`
        DB       6
        FCC     'DIGIT?'
DIGTQ
        JSR     TOR
        JSR     DOLIT
        DW      '0'
        JSR     SUBB
        JSR     DOLIT
        DW      9
        JSR     OVER
        JSR     LESS
        JSR     QBRAN
        DW      DGTQ1
        JSR     DOLIT
        DW      7
        JSR     SUBB
        JSR     DUPP
        JSR     DOLIT
        DW      10
        JSR     LESS
        JSR     ORR
DGTQ1:  JSR     DUPP
        JSR     RFROM
        JMP     ULESS

;       NUMBER? ( a -- n T | a F )
;       Convert a number string to
;         integer. Push a flag on tos.

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'NUMBER?'
NUMBQ
        JSR     BASE
        JSR     AT
        JSR     TOR
        JSR     ZERO
        JSR     OVER
        JSR     COUNT
        JSR     OVER
        JSR     CAT
        JSR     DOLIT
        DW      '$'
        JSR     EQUAL
        JSR     QBRAN
        DW      NUMQ1
        JSR     HEX
        JSR     SWAP
        JSR     ONEP
        JSR     SWAP
        JSR     ONEM
NUMQ1:  JSR     OVER
        JSR     CAT
        JSR     DOLIT
        DW      '-'
        JSR     EQUAL
        JSR     TOR
        JSR     SWAP
        JSR     RAT
        JSR     SUBB
        JSR     SWAP
        JSR     RAT
        JSR     PLUS
        JSR     QDUP
        JSR     QBRAN
        DW      NUMQ6
        JSR     ONEM
        JSR     TOR
NUMQ2:  JSR     DUPP
        JSR     TOR
        JSR     CAT
        JSR     BASE
        JSR     AT
        JSR     DIGTQ
        JSR     QBRAN
        DW      NUMQ4
        JSR     SWAP
        JSR     BASE
        JSR     AT
        JSR     STAR
        JSR     PLUS
        JSR     RFROM
        JSR     ONEP
        JSR     DONXT
        DW      NUMQ2
        JSR     RAT
        JSR     SWAP
        JSR     DROP
        JSR     QBRAN
        DW      NUMQ3
        JSR     NEGAT
NUMQ3:  JSR     SWAP
        BRA     NUMQ5
NUMQ4:  JSR     RFROM
        JSR     RFROM
        JSR     DDROP
        JSR     DDROP
        JSR     ZERO
NUMQ5:  JSR     DUPP
NUMQ6:  JSR     RFROM
        JSR     DDROP
        JSR     RFROM
        JSR     BASE
        JMP     STORE

;; Basic I/O

;       KEY     ( -- c )
;       Wait for and return an
;         input character.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'KEY'
KEY
KEY1:   JSR     QKEY
        JSR     QBRAN
        DW      KEY1
        RTS

;       NUF?    ( -- t )
;       Return false if no input,
;         else pause and if CR return true.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'NUF?'
NUFQ
        JSR     QKEY
        JSR     DUPP
        JSR     QBRAN
        DW      NUFQ1
        JSR     DDROP
        JSR     KEY
        JSR     DOLIT
        DW      CRR
        JMP     EQUAL
NUFQ1:  RTS

;       SPACE   ( -- )
;         Send  blank character to
;         output device.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'SPACE'
SPACE
        JSR     BLANK
        JMP     EMIT

;       SPACES  ( +n -- )
;         Send n spaces to output device.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'SPACES'
SPACS
        JSR     ZERO
        JSR     MAX
        JSR     TOR
        BRA     CHAR2
CHAR1:  JSR     SPACE
CHAR2:  JSR     DONXT
        DW      CHAR1
        RTS

;       TYPE    ( b u -- )
;       Output u characters from b.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'TYPE'
TYPES
        JSR     TOR
        BRA     TYPE2
TYPE1:  JSR     DUPP
        JSR     CAT
        JSR     EMIT
        JSR     ONEP
TYPE2:  JSR     DONXT
        DW      TYPE1
        JMP     DROP

;       CR      ( -- )
;       Output a carriage return
;         and a line feed.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'CR'
CR
        JSR     DOLIT
        DW      CRR
        JSR     EMIT
        JSR     DOLIT
        DW      LF
        JMP     EMIT

;       do$     ( -- a )
;         Return  address of a compiled
;         string.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+3
        FCC     'do$'
DOSTR
        JSR     RFROM
        JSR     RAT
        JSR     RFROM
        JSR     COUNT
        JSR     PLUS
        JSR     TOR
        JSR     SWAP
        JSR     TOR
        RTS

;       $"|     ( -- a )
;       Run time routine compiled by $".
;         Return address of a compiled string.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+3
        FCC     '$"|'
STRQP
        JSR     DOSTR
        RTS

;       ."|     ( -- )
;       Run time routine of ." .
;         Output a compiled string.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+3
        FCC     '."|'
DOTQP
        JSR     DOSTR
        JSR     COUNT
        JMP     TYPES

;       .R      ( n +n -- )
;       Display an integer in a field
;         of n columns, right justified.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '.R'
DOTR
        JSR     TOR
        JSR     STR
        JSR     RFROM
        JSR     OVER
        JSR     SUBB
        JSR     SPACS
        JMP     TYPES

;       U.R     ( u +n -- )
;       Display an unsigned integer
;         in n column, right justified.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'U.R'
UDOTR
        JSR     TOR
        JSR     BDIGS
        JSR     DIGS
        JSR     EDIGS
        JSR     RFROM
        JSR     OVER
        JSR     SUBB
        JSR     SPACS
        JMP     TYPES

;       U.      ( u -- )
;       Display an unsigned integer
;         in free format.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'U.'
UDOT
        JSR     BDIGS
        JSR     DIGS
        JSR     EDIGS
        JSR     SPACE
        JMP     TYPES

;       .       ( w -- )
;       Display an integer in free
;         format, preceeded by a space.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '.'
DOT
        JSR     BASE
        JSR     AT
        JSR     DOLIT
        DW      10
        JSR     XORR    ;?decimal
        JSR     QBRAN
        DW      DOT1
        JMP     UDOT
DOT1:   JSR     STR
        JSR     SPACE
        JMP     TYPES

;       ?       ( a -- )
;         Display contents in memory cell.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     '?'
QUEST
        JSR     AT
        JMP     DOT

;; Parsing

;       parse   ( b u c -- b u delta ; <string> )
;       Scan string delimited by c.
;         Return found string and its offset.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'parse'
PARS
        JSR     TEMP
        JSR     STORE
        JSR     OVER
        JSR     TOR
        JSR     DUPP
        JSR     QBRAN
        DW      PARS8
        JSR     ONEM
        JSR     TEMP
        JSR     AT
        JSR     BLANK
        JSR     EQUAL
        JSR     QBRAN
        DW      PARS3
        JSR     TOR
PARS1:  JSR     BLANK
        JSR     OVER
        JSR     CAT     ;skip leading blanks ONLY
        JSR     SUBB
        JSR     ZLESS
        JSR     INVER
        JSR     QBRAN
        DW      PARS2
        JSR     ONEP
        JSR     DONXT
        DW      PARS1
        JSR     RFROM
        JSR     DROP
        JSR     ZERO
        JMP     DUPP
PARS2:  JSR     RFROM
PARS3:  JSR     OVER
        JSR     SWAP
        JSR     TOR
PARS4:  JSR     TEMP
        JSR     AT
        JSR     OVER
        JSR     CAT
        JSR     SUBB    ;scan for delimiter
        JSR     TEMP
        JSR     AT
        JSR     BLANK
        JSR     EQUAL
        JSR     QBRAN
        DW      PARS5
        JSR     ZLESS
PARS5:  JSR     QBRAN
        DW      PARS6
        JSR     ONEP
        JSR     DONXT
        DW      PARS4
        JSR     DUPP
        JSR     TOR
        BRA     PARS7
PARS6:  JSR     RFROM
        JSR     DROP
        JSR     DUPP
        JSR     ONEP
        JSR     TOR
PARS7:  JSR     OVER
        JSR     SUBB
        JSR     RFROM
        JSR     RFROM
        JMP     SUBB
PARS8:  JSR     OVER
        JSR     RFROM
        JMP     SUBB

;       PARSE   ( c -- b u ; <string> )
;       Scan input stream and return
;         counted string delimited by c.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PARSE'
PARSE
        JSR     TOR
        JSR     TIB
        JSR     INN
        JSR     AT
        JSR     PLUS    ;current input buffer pointer
        JSR     NTIB
        JSR     AT
        JSR     INN
        JSR     AT
        JSR     SUBB    ;remaining count
        JSR     RFROM
        JSR     PARS
        JSR     INN
        JMP     PSTOR

;       .(      ( -- )
;       Output following string up to next ) .

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+2
        FCC     '.('
DOTPR
        JSR     DOLIT
        DW      ')'
        JSR     PARSE
        JMP     TYPES

;       (       ( -- )
;         Ignore following string up to next ).
;         A comment.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+1
        FCC     '('
PAREN
        JSR     DOLIT
        DW      ')'
        JSR     PARSE
        JMP     DDROP

;       \       ( -- )
;         Ignore following text till
;         end of line.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+1
        FCC     '\'
BKSLA
        JSR     NTIB
        JSR     AT
        JSR     INN
        JMP     STORE

;       WORD    ( c -- a ; <string> )
;       Parse a word from input stream
;         and copy it to code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'WORD'
WORDD
        JSR     PARSE
        JSR     HERE
        JSR     CELLP
        JMP     PACKS

;       TOKEN   ( -- a ; <string> )
;       Parse a word from input stream
;         and copy it to name dictionary.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'TOKEN'
TOKEN
        JSR     BLANK
        JMP     WORDD

;; Dictionary search

;       NAME>   ( na -- ca )
;       Return a code address given
;         a name address.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'NAME>'
NAMET
        JSR     COUNT
        JSR     DOLIT
        DW      31
        JSR     ANDD
        JMP     PLUS

;       SAME?   ( a a u -- a a f \ -0+ )
;       Compare u cells in two
;         strings. Return 0 if identical.

        DW      LINK`
        LOC
LINK`
        DB       5
        FCC     'SAME?'
SAMEQ
        JSR     ONEM
        JSR     TOR
        BRA     SAME2
SAME1:  JSR     OVER
        JSR     RAT
        JSR     PLUS
        JSR     CAT
        JSR     OVER
        JSR     RAT
        JSR     PLUS
        JSR     CAT
        JSR     SUBB
        JSR     QDUP
        JSR     QBRAN
        DW      SAME2
        JSR     RFROM
        JMP     DROP
SAME2:  JSR     DONXT
        DW      SAME1
        JMP     ZERO

;       find    ( a va -- ca na | a F )
;         Search vocabulary for string.
;         Return ca and na if succeeded.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'find'
FIND
        JSR     SWAP
        JSR     DUPP
        JSR     CAT
        JSR     TEMP
        JSR     STORE
        JSR     DUPP
        JSR     AT
        JSR     TOR
        JSR     CELLP
        JSR     SWAP
FIND1:  JSR     AT
        JSR     DUPP
        JSR     QBRAN
        DW      FIND6
        JSR     DUPP
        JSR     AT
        JSR     DOLIT
        DW      MASKK
        JSR     ANDD
        JSR     RAT
        JSR     XORR
        JSR     QBRAN
        DW      FIND2
        JSR     CELLP
        JSR     DOLIT
        DW      -1
        BRA     FIND3
FIND2:  JSR     CELLP
        JSR     TEMP
        JSR     AT
        JSR     SAMEQ
FIND3:  JSR     BRAN
        DW      FIND4
FIND6:  JSR     RFROM
        JSR     DROP
        JSR     SWAP
        JSR     CELLM
        JMP     SWAP
FIND4:  JSR     QBRAN
        DW      FIND5
        JSR     CELLM
        JSR     CELLM
        BRA     FIND1
FIND5:  JSR     RFROM
        JSR     DROP
        JSR     SWAP
        JSR     DROP
        JSR     CELLM
        JSR     DUPP
        JSR     NAMET
        JMP     SWAP

;       NAME?   ( a -- ca na | a F )
;         Search vocabularies for a string.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'NAME?'
NAMEQ
        JSR     CNTXT
        JMP     FIND

;; Terminal response

;       ^H      ( bot eot cur -- bot eot cur )
;         Backup cursor by one character.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '^h'
BKSP
        JSR     TOR
        JSR     OVER
        JSR     RFROM
        JSR     SWAP
        JSR     OVER
        JSR     XORR
        JSR     QBRAN
        DW      BACK1
        JSR     DOLIT
        DW      BKSPP
        JSR     EMIT
        JSR     ONEM
        JSR     BLANK
        JSR     EMIT
        JSR     DOLIT
        DW      BKSPP
        JMP     EMIT
BACK1:  RTS

;       TAP     ( bot eot cur c -- bot eot cur )
;         Accept and echo key stroke
;         and bump cursor.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'TAP'
TAP
        JSR     DUPP
        JSR     EMIT
        JSR     OVER
        JSR     CSTOR
        JMP     ONEP

;       kTAP    ( bot eot cur c -- bot eot cur )
;       Process a key stroke,
;         CR or backspace.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'kTAP'
KTAP
        JSR     DUPP
        JSR     DOLIT
        DW      CRR
        JSR     XORR
        JSR     QBRAN
        DW      KTAP2
        JSR     DOLIT
        DW      BKSPP
        JSR     XORR
        JSR     QBRAN
        DW      KTAP1
        JSR     BLANK
        JMP     TAP
KTAP1:  JMP     BKSP
KTAP2:  JSR     DROP
        JSR     SWAP
        JSR     DROP
        JMP     DUPP

;       accept  ( b u -- b u )
;       Accept characters to input
;         buffer. Return with actual count.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'ACCEPT'
ACCEP
        JSR     OVER
        JSR     PLUS
        JSR     OVER
ACCP1:  JSR     DDUP
        JSR     XORR
        JSR     QBRAN
        DW      ACCP4
        JSR     KEY
        JSR     DUPP
        JSR     BLANK
        JSR     DOLIT
        DW      127
        JSR     WITHI
        JSR     QBRAN
        DW      ACCP2
        JSR     TAP
        BRA     ACCP3
ACCP2:  JSR     KTAP
ACCP3:  BRA     ACCP1
ACCP4:  JSR     DROP
        JSR     OVER
        JMP     SUBB

;       QUERY   ( -- )
;       Accept input stream to
;         terminal input buffer.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'QUERY'
QUERY
        JSR     TIB
        JSR     DOLIT
        DW      80
        JSR     ACCEP
        JSR     NTIB
        JSR     STORE
        JSR     DROP
        JSR     ZERO
        JSR     INN
        JMP     STORE

;       ABORT   ( -- )
;       Reset data stack and
;         jump to QUIT.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'ABORT'
ABORT
        JSR     PRESE
        JMP     QUIT

;       abort"  ( f -- )
;         Run time routine of ABORT".
;         Abort with a message.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+6
        FCC     'abort"'
ABORQ
        JSR     QBRAN
        DW      ABOR2   ;text flag
        JSR     DOSTR
ABOR1:  JSR     SPACE
        JSR     COUNT
        JSR     TYPES
        JSR     DOLIT
        DW      '?'
        JSR     EMIT
        JSR     CR
        JMP     ABORT   ;pass error string
ABOR2:  JSR     DOSTR
        JMP     DROP

;; The text interpreter

;       $INTERPRET      ( a -- )
;       Interpret a word. If failed,
;         try to convert it to an integer.

        DW      LINK`
        LOC
LINK`
        DB      10
        FCC     '$INTERPRET'
INTER
        JSR     NAMEQ
        JSR     QDUP    ;?defined
        JSR     QBRAN
        DW      INTE1
        JSR     AT
        JSR     DOLIT
        DW       COMPO*256
        JSR     ANDD    ;?compile only lexicon bits
        JSR     ABORQ
        DB      13
        FCC     ' compile only'
        JMP     EXECU
INTE1:  JSR     NUMBQ   ;convert a number
        JSR     QBRAN
        DW      ABOR1
        RTS

;       [       ( -- )
;         Start  text interpreter.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+1
        FCC     '['
LBRAC
        JSR     DOLIT
        DW      INTER
        JSR     TEVAL
        JMP     STORE

;       .OK     ( -- )
;         Display 'ok' while interpreting.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     '.OK'
DOTOK
        JSR     DOLIT
        DW      INTER
        JSR     TEVAL
        JSR     AT
        JSR     EQUAL
        JSR     QBRAN
        DW      DOTO1
        JSR     DOTQP
        DB      3
        FCC     ' ok'
DOTO1:  JMP     CR

;       ?STACK  ( -- )
;         Abort if stack underflows.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     '?STACK'
QSTAC
        JSR     DEPTH
        JSR     ZLESS   ;check only for underflow
        JSR     ABORQ
        DB      11
        FCC     ' underflow '
        RTS

;       EVAL    ( -- )
;         Interpret  input stream.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'EVAL'
EVAL
EVAL1:  JSR     TOKEN
        JSR     DUPP
        JSR     CAT     ;?input stream empty
        JSR     QBRAN
        DW      EVAL2
        JSR     TEVAL
        JSR     ATEXE
        JSR     QSTAC   ;evaluate input, check stack
        JSR     BRAN
        DW      EVAL1
EVAL2:  JSR     DROP
        JMP     DOTOK

;       PRESET  ( -- )
;         Reset data stack pointer and
;         terminal input buffer.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'PRESET'
PRESE
        JSR     DOLIT
        DW      SPP
        JSR     SPSTO
        JSR     DOLIT
        DW      TIBB
        JSR     NTIB
        JSR     CELLP
        JMP     STORE

;       QUIT    ( -- )
;       Reset return stack pointer
;         and start text interpreter.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'QUIT'
QUIT
        JSR     DOLIT
        DW      RPP
        JSR     RPSTO   ;reset return stack pointer
QUIT1:  JSR     LBRAC   ;start interpretation
QUIT2:  JSR     QUERY   ;get input
        JSR     EVAL
        BRA     QUIT2   ;continue till error

;; The compiler

;       '       ( -- ca )
;         Search vocabularies for
;         next word in input stream.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     "'"
TICK
        JSR     TOKEN
        JSR     NAMEQ   ;?defined
        JSR     QBRAN
        DW      ABOR1
        RTS     ;yes, push code address

;       ALLOT   ( n -- )
;         Allocate n bytes to  code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'ALLOT'
ALLOT
        JSR     CP
        JMP     PSTOR

;       ,       ( w -- )
;         Compile an integer into
;         code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     ','
COMMA
        JSR     HERE
        JSR     DUPP
        JSR     CELLP   ;cell boundary
        JSR     CP
        JSR     STORE
        JSR     STORE
        RTS     

;       C,      ( c -- )
;         Compile a byte into
;         code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'C,'
CCOMMA
        JSR     HERE
        JSR     DUPP
        JSR     ONEP
        JSR     CP
        JSR     STORE
        JSR     CSTOR
        RTS     

;       [COMPILE]       ( -- ; <string> )
;         Compile next immediate
;         word into code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+9
        FCC     '[COMPILE]'
BCOMP
        JSR     TICK
        JMP     JSRC

;       COMPILE ( -- )
;         Compile next jsr in
;         colon list to code dictionary.

        DW      LINK`
        LOC
LINK`
        DB      COMPO+7
        FCC     'COMPILE'
COMPI
        JSR     RFROM
        JSR     ONEP
        JSR     DUPP
        JSR     AT
        JSR     JSRC    ;compile subroutine
        JSR     CELLP
        JMP     TOR

;       LITERAL ( w -- )
;         Compile tos to dictionary
;         as an integer literal.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+7
        FCC     'LITERAL'
LITER
        JSR     COMPI
        JSR     DOLIT
        JMP     COMMA

;       $,"     ( -- )
;       Compile a literal string
;         up to next " .

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     '$,"'
STRCQ
        JSR     DOLIT
        DW      '"'
        JSR     PARSE
        JSR     HERE
        JSR     PACKS   ;string to code dictionary
        JSR     COUNT
        JSR     PLUS    ;calculate aligned end of string
        JSR     CP
        JMP     STORE

;; Structures

;       FOR     ( -- a )
;       Start a FOR-NEXT loop
;         structure in a colon definition.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+3
        FCC     'FOR'
FOR
        JSR     COMPI
        JSR     TOR
        JMP     HERE

;       NEXT    ( a -- )
;         Terminate a FOR-NEXT loop.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+4
        FCC     'NEXT'
NEXT
        JSR     COMPI
        JSR     DONXT
        JMP     COMMA

;       BEGIN   ( -- a )
;       Start an infinite or
;         indefinite loop structure.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+5
        FCC     'BEGIN'
BEGIN
        JMP     HERE

;       UNTIL   ( a -- )
;       Terminate a BEGIN-UNTIL
;         indefinite loop structure.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+5
        FCC     'UNTIL'
UNTIL
        JSR     COMPI
        JSR     QBRAN
        JMP     COMMA

;       AGAIN   ( a -- )
;       Terminate a BEGIN-AGAIN
;         infinite loop structure.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+5
        FCC     'AGAIN'
AGAIN
        JSR     COMPI
        JSR     BRAN
        JMP     COMMA

;       IF      ( -- A )
;         Begin a conditional branch.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+2
        FCC     'IF'
IFF
        JSR     COMPI
        JSR     QBRAN
        JSR     HERE
        JSR     ZERO
        JMP     COMMA

;   THEN        ( A -- )
;               Terminate a conditional branch structure.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+4
        FCC     'THEN'
THENN
        JSR     HERE
        JSR     SWAP
        JMP     STORE

;   ELSE        ( A -- A )
;               Start the false clause in an IF-ELSE-THEN structure.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+4
        FCC     'ELSE'
ELSEE
        JSR     COMPI
        JSR     BRAN
        JSR     HERE
        JSR     ZERO
        JSR     COMMA
        JSR     SWAP
        JSR     HERE
        JSR     SWAP
        JMP     STORE

;   AHEAD       ( -- A )
;               Compile a forward branch instruction.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+5
        FCC     'AHEAD'
AHEAD
        JSR     COMPI
        JSR     BRAN
        JSR     HERE
        JSR     ZERO
        JMP     COMMA

;   WHILE       ( a -- A a )
;       Conditional branch out of a BEGIN-WHILE-REPEAT loop.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+5
        FCC     'WHILE'
WHILE
        JSR     COMPI
        JSR     QBRAN
        JSR     HERE
        JSR     ZERO
        JSR     COMMA
        JMP     SWAP

;   REPEAT      ( A a -- )
;               Terminate a BEGIN-WHILE-REPEAT indefinite loop.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+6
        FCC     'REPEAT'
REPEA
        JSR     COMPI
        JSR     BRAN
        JSR     COMMA
        JSR     HERE
        JSR     SWAP
        JMP     STORE

;   AFT         ( a -- a A )
;               Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+3
        FCC     'AFT'
AFT
        JSR     DROP
        JSR     AHEAD
        JSR     HERE
        JMP     SWAP

;   ABORT"      ( -- ; <string> )
;               Conditional abort with an error message.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+6
        FCC     'ABORT"'
ABRTQ
        JSR     COMPI
        JSR     ABORQ
        JMP     STRCQ

;   $"          ( -- ; <string> )
;               Compile an inline string literal.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+2
        FCC     '$"'
STRQ
        JSR     COMPI
        JSR     STRQP
        JMP     STRCQ

;   ."          ( -- ; <string> )
;               Compile an inline string literal to be typed out at run time.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+2
        FCC     '."'
DOTQ
        JSR     COMPI
        JSR     DOTQP
        JMP     STRCQ

;; Name compiler

;       ?UNIQUE ( a -- a )
;         Display a warning message
;         if word already exists.

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     '?UNIQUE'
UNIQU
        JSR     DUPP
        JSR     NAMEQ   ;?name exists
        JSR     QBRAN
        DW      UNIQ1
        JSR     DOTQP   ;redef are OK
        DB       7
        FCC     ' reDef '       
        JSR     OVER
        JSR     COUNT
        JSR     TYPES   ;just in case
UNIQ1:  JMP     DROP

;       $,n     ( na -- )
;         Build a new dictionary name
;         using string at na.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     '$,n'
SNAME
        JSR     DUPP
        JSR     CAT     ;?null input
        JSR     QBRAN
        DW      PNAM1
        JSR     UNIQU   ;?redefinition
        JSR     DUPP
        JSR     COUNT
        JSR     PLUS
        JSR     CP
        JSR     STORE
        JSR     DUPP
        JSR     LAST
        JSR     STORE   ;save na for vocabulary link
        JSR     CELLM   ;link address
        JSR     CNTXT
        JSR     AT
        JSR     SWAP
        JSR     STORE
        RTS     ;save code pointer
PNAM1:  JSR     STRQP
        DB      5
        FCC     ' name' ;null input
        JMP     ABOR1

;; FORTH compiler

;       $COMPILE        ( a -- )
;         Compile next word to
;         dictionary as a token or literal.

        DW      LINK`
        LOC
LINK`
        DB      8
        FCC     '$COMPILE'
SCOMP
        JSR     NAMEQ
        JSR     QDUP    ;?defined
        JSR     QBRAN
        DW      SCOM2
        JSR     AT
        JSR     DOLIT
        DW       IMEDD*256
        JSR     ANDD    ;?immediate
        JSR     QBRAN
        DW      SCOM1
        JMP     EXECU
SCOM1:  JMP     JSRC
SCOM2:  JSR     NUMBQ   ;try to convert to number
        JSR     QBRAN
        DW      ABOR1
        JMP     LITER

;       OVERT   ( -- )
;         Link a new word into vocabulary.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'OVERT'
OVERT
        JSR     LAST
        JSR     AT
        JSR     CNTXT
        JMP     STORE

;       ;       ( -- )
;       Terminate a colon definition.

        DW      LINK`
        LOC
LINK`
        DB      IMEDD+COMPO+1
        FCC     ';'
SEMIS
        JSR     COMPI
        JSR     EXIT
        JSR     LBRAC
        JMP     OVERT

;       ]       ( -- )
;         Start compiling words in
;         input stream.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     ']'
RBRAC
        JSR     DOLIT
        DW      SCOMP
        JSR     TEVAL
        JMP     STORE

;       JSR,    ( ca -- )
;         Compile a subroutine call.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'JSR,'
JSRC
        JSR     DOLIT
        DW      $16     ;JSR
        JSR     CCOMMA
        JMP     COMMA

;       :       ( -- ; <string> )
;       Start a new colon definition
;         using next word as its name.

        DW      LINK`
        LOC
LINK`
        DB      1
        FCC     ':'
COLON
        JSR     TOKEN
        JSR     SNAME
        JMP     RBRAC

;       IMMEDIATE       ( -- )
;         Make last compiled word
;         an immediate word.

        DW      LINK`
        LOC
LINK`
        DB      9
        FCC     'IMMEDIATE'
IMMED
        JSR     DOLIT
        DW       IMEDD*256
        JSR     LAST
        JSR     AT
        JSR     AT
        JSR     ORR
        JSR     LAST
        JSR     AT
        JMP     STORE

;; Defining words

;       CREATE  ( -- ; <string> )
;         Compile a new array
;         without allocating space.

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'CREATE'
CREAT
        JSR     TOKEN
        JSR     SNAME
        JSR     OVERT
        JSR     COMPI
        JSR     DOVAR
        RTS

;       VARIABLE        ( -- ; <string> )
;       Compile a new variable
;         initialized to 0.

        DW      LINK`
        LOC
LINK`
        DB      8
        FCC     'VARIABLE'
VARIA
        JSR     CREAT
        JSR     ZERO
        JMP     COMMA

;; Tools

;       _TYPE   ( b u -- )
;       Display a string. Filter
;         non-printing characters.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     '_TYPE'
UTYPE
        JSR     TOR     ;start count down loop
        BRA     UTYP2   ;skip first pass
UTYP1:  JSR     DUPP
        JSR     CAT
        JSR     TCHAR
        JSR     EMIT    ;display only printable
        JSR     ONEP    ;increment address
UTYP2:  JSR     DONXT
        DW      UTYP1   ;loop till done
        JMP     DROP

;       dm+     ( a u -- a )
;         Dump u bytes from ,
;         leaving a+u on  stack.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'dm+'
DUMPP
        JSR     OVER
        JSR     DOLIT
        DW      4
        JSR     UDOTR   ;display address
        JSR     SPACE
        JSR     TOR     ;start count down loop
        BRA     PDUM2   ;skip first pass
PDUM1:  JSR     DUPP
        JSR     CAT
        JSR     DOLIT
        DW      3
        JSR     UDOTR   ;display numeric data
        JSR     ONEP    ;increment address
PDUM2:  JSR     DONXT
        DW      PDUM1   ;loop till done
        RTS

;       DUMP    ( a u -- )
;       Dump u bytes from a,
;         in a formatted manner.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DUMP'
DUMP
        JSR     BASE
        JSR     AT
        JSR     TOR
        JSR     HEX     ;save radix, set hex
        JSR     DOLIT
        DW      16
        JSR     SLASH   ;change count to lines
        JSR     TOR     ;start count down loop
DUMP1:  JSR     CR
        JSR     DOLIT
        DW      16
        JSR     DDUP
        JSR     DUMPP   ;display numeric
        JSR     ROT
        JSR     ROT
        JSR     SPACE
        JSR     SPACE
        JSR     UTYPE   ;display printable characters
        JSR     NUFQ
        JSR     INVER   ;user control
        JSR     QBRAN
        DW      DUMP2
        JSR     DONXT
        DW      DUMP1   ;loop till done
        BRA     DUMP3
DUMP2:  JSR     RFROM
        JSR     DROP    ;cleanup loop stack
DUMP3:  JSR     DROP
        JSR     RFROM
        JSR     BASE
        JMP     STORE   ;restore radix

;       .S      ( ... -- ... )
;         Display  contents of stack.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     '.S'
DOTS
        JSR     CR
        JSR     DEPTH   ;stack depth
        JSR     TOR     ;start count down loop
        BRA     DOTS2   ;skip first pass
DOTS1:  JSR     RAT
        JSR     PICK
        JSR     DOT     ;index stack, display contents
DOTS2:  JSR     DONXT
        DW      DOTS1   ;loop till done
        JSR     DOTQP
        DB      5
        FCC     ' <sp '
        RTS

;       >NAME   ( ca -- na | F )
;       Convert code address
;         to a name address.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     '>NAME'
TNAME
        JSR     CNTXT   ;vocabulary link
TNAM2:  JSR     AT
        JSR     DUPP    ;?last word in a vocabulary
        JSR     QBRAN
        DW      TNAM4
        JSR     DDUP
        JSR     NAMET
        JSR     XORR    ;compare
        JSR     QBRAN
        DW      TNAM3
        JSR     CELLM   ;continue with next word
        BRA     TNAM2
TNAM3:  JSR     SWAP
        JMP     DROP
TNAM4:  JSR     DDROP
        JMP     ZERO

;       .ID     ( na -- )
;         Display  name at address.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     '.ID'
DOTID
        JSR     QDUP    ;if zero no name
        JSR     QBRAN
        DW      DOTI1
        JSR     COUNT
        JSR     DOLIT
        DW      $1F
        JSR     ANDD    ;mask lexicon bits
        JMP     UTYPE
DOTI1:  JSR     DOTQP
        DB      9
        FCC     ' {noName}'
        RTS

;       SEE     ( -- ; <string> )
;       A simple decompiler.
;         Updated for byte machines.

        DW      LINK`
        LOC
LINK`
        DB      3
        FCC     'SEE'
SEE
        JSR     TICK    ;starting address
        JSR     CR
        JSR     ONEM
SEE1:   JSR     ONEP
        JSR     DUPP
        JSR     AT
        JSR     DUPP    ;?does it contain a zero
        JSR     QBRAN
        DW      SEE2
        JSR     TNAME   ;?is it a name
SEE2:   JSR     QDUP    ;name address or zero
        JSR     QBRAN
        DW      SEE3
        JSR     SPACE
        JSR     DOTID   ;display name
        JSR     ONEP
        BRA     SEE4
SEE3:   JSR     DUPP
        JSR     CAT
        JSR     UDOT    ;display number
SEE4:   JSR     NUFQ    ;user control
        JSR     QBRAN
        DW      SEE1
        JMP     DROP

;       WORDS   ( -- )
;         Display names in vocabulary.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'WORDS'
WORDS
        JSR     CR
        JSR     CNTXT   ;only in context
WORS1:  JSR     AT
        JSR     QDUP    ;?at end of list
        JSR     QBRAN
        DW      WORS2
        JSR     DUPP
        JSR     SPACE
        JSR     DOTID   ;display a name
        JSR     CELLM
        JSR     NUFQ    ;user control
        JSR     QBRAN
        DW      WORS1
        JSR     DROP
WORS2:  RTS

        
;; Hardware reset

;       hi      ( -- )
;         Display sign-on message.

        DW      LINK`
        LOC
LINK`
        DB      2
        FCC     'hi'
HI
        JSR     CR
        JSR     DOTQP   ;initialize I/O
        DB      13
        FCC     '12eForth v'
        DB      VER+'0'
        DB      '.'
        DB      EXT+'0' ;version
        JMP     CR

;       'BOOT   ( -- a )
;       The application startup vector.

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     "'BOOT"
TBOOT
        JSR     DOVAR
        DW      HI       ;application to boot

;       COLD    ( -- )
;       The hilevel cold start sequence.

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'COLD'
COLD
COLD1:  JSR     DOLIT
        DW      UZERO
        JSR     DOLIT
        DW      UPP
        JSR     DOLIT
        DW      ULAST-UZERO
        JSR     CMOVE   ;initialize user area
        JSR     PRESE   ;initialize data stack and TIB
        JSR     TBOOT
        JSR     ATEXE   ;application boot
        JSR     OVERT
        JMP     QUIT    ;start interpretation

;==============================================
;       I/O Ports

;       PORTA

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PORTA'
PORTA   MOVW    #$0,2,-X
        RTS

;       PORT

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PORTB'
PORTB   MOVW    #$1,2,-X
        RTS

;       DDRA

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DDRA'
DDRA    MOVW    #$2,2,-X
        RTS

;       DDRB

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DDRB'
DDRB    MOVW    #$3,2,-X
        RTS

;       PORTE

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PORTE'
PORTE   MOVW    #$8,2,-X
        RTS

;       DDRE

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DDRE'
DDRE    MOVW    #$9,2,-X
        RTS

;       PWCLK

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWCLK'
PWCLK   MOVW    #$40,2,-X
        RTS

;       PWPOL

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWPOL'
PWPOL   MOVW    #$41,2,-X
        RTS

;       PWEN

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'PWEN'
PWEN    MOVW    #$42,2,-X
        RTS

;       PWSCAL0

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'PWSCAL0'
PSCL0   MOVW    #$44,2,-X
        RTS

;       PWSCNT0

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'PWSCNT0'
PSCN0   MOVW    #$45,2,-X
        RTS

;       PWSCAL1

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'PWSCAL1'
PSCL1   MOVW    #$46,2,-X
        RTS

;       PWSCNT1

        DW      LINK`
        LOC
LINK`
        DB      7
        FCC     'PWSCNT1'
PSCN1   MOVW    #$47,2,-X
        RTS

;       PWCNT0

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWCNT0'
PCNT0   MOVW    #$48,2,-X
        RTS

;       PWCNT1

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWCNT1'
PCNT1   MOVW    #$49,2,-X
        RTS

;       PWCNT2

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWCNT2'
PCNT2   MOVW    #$4A,2,-X
        RTS

;       PWCNT3

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWCNT3'
PCNT3   MOVW    #$4B,2,-X
        RTS

;       PWPER0

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWPER0'
PPER0   MOVW    #$4C,2,-X
        RTS

;       PWPER1

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWPER1'
PPER1   MOVW    #$4D,2,-X
        RTS

;       PWPER2

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWPER2'
PPER2   MOVW    #$4E,2,-X
        RTS

;       PWPER3

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWPER3'
PPER3   MOVW    #$4F,2,-X
        RTS

;       PWDTY0

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWDTY0'
PDTY0   MOVW    #$50,2,-X
        RTS

;       PWDTY1

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWDTY1'
PDTY1   MOVW    #$51,2,-X
        RTS

;       PWDTY2

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWDTY2'
PDTY2   MOVW    #$52,2,-X
        RTS

;       PWDTY3

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWDTY3'
PDTY3   MOVW    #$53,2,-X
        RTS

;       PWCTL

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PWCTL'
PWCTL   MOVW    #$54,2,-X
        RTS

;       PORTP

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PORTP'
PORTP   MOVW    #$56,2,-X
        RTS

;       DDRP

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DDRP'
DDRP    MOVW    #$57,2,-X
        RTS

;       PORTAD

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'PORTAD'
PORTAD  MOVW    #$6F,2,-X
        RTS

;       PORTT

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PORTT'
PORTT   MOVW    #$AE,2,-X
        RTS

;       DDRT

        DW      LINK`
        LOC
LINK`
        DB      4
        FCC     'DDRT'
DDRT    MOVW    #$AF,2,-X
        RTS

;       SP0CR1

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'SP0CR1'
SPCR1   MOVW    #$D0,2,-X
        RTS

;       SP0CR2

        DW      LINK`
        LOC
LINK`
        DB      6
        FCC     'SP0CR2'
SPCR2   MOVW    #$D1,2,-X
        RTS

;       SP0BR

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'SP0BR'
SP0BR   MOVW    #$D2,2,-X
        RTS

;       SP0SR

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'SP0SR'
SP0SR   MOVW    #$D3,2,-X
        RTS

;       SP0DR

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'SP0DR'
SP0DR   MOVW    #$D5,2,-X
        RTS

;       PORTS

        DW      LINK`
        LOC
LINK`
        DB      5
        FCC     'PORTS'
PORTS   MOVW    #$D6,2,-X
        RTS

;       DDRS

	DW	LINK`
	LOC
LINK`
	DB	4
	FCC	'DDRS'
DDRS	MOVW	#$D7,2,-X
	RTS

;	PORTDLC

	DW	LINK`
	LOC
LINK`
	DB	7
	FCC	'PORTDLC'
PDLC	MOVW	#$FE,2,-X
	RTS

;	DDRDLC

	DW	LINK`
	LOC
LINK`
	DB	6
	FCC	'DDRDLC'
DDLC	MOVW	#$FF,2,-X
	RTS

;==============================================
;	A/D conversion

;	ADCONV.ON

	DW	LINK`
	LOC
LINK`
	DB	9
	FCC	'ADCONV.ON'
ADON	JSR	DOLIT
	DW	$0C0
ADON1	JSR	DOLIT
	DW	$62	;ATDCTL2
	JMP	CSTOR

;	ADCONV.OFF

	DW	LINK`
	LOC
LINK`
	DB	10
	FCC	'ADCONV.OFF'
ADOFF	JSR	ZERO
	JMP	ADON1

;	WAIT.FOR.CONV

	DW	LINK`
	LOC
LINK`
	DB	13
	FCC	'WAIT.FOR.CONV'
WADC	JSR	DOLIT
	DW	7
	JSR	DOLIT
	DW	$66	;ATDSTAT
	JSR	QHI
	JSR	QBRAN
	DW	WADC
	RTS

;	ADCONV ( ch# -- val )

	DW	LINK`
	LOC
LINK`
	DB	6
	FCC	'ADCONV'
CONV	JSR	ADON
	JSR	DUPP
	JSR	DOLIT
	DW	$65	;ATDCTL5
	JSR	CSTOR
	JSR	WADC
	JSR	DOLIT
	DW	7
	JSR	ANDD
	JSR	CELLS
	JSR	DOLIT
	DW	$70	;ADR0H
	JSR	PLUS
	JSR	AT
	JMP	ADOFF

;	GET.4VALUES ( -- n n n n )

	DW	LINK`
	LOC
LINK`
	DB	11
	FCC	'GET.4VALUES'
GET4	JSR	DOLIT
	DW	$70	;ADR0H
	JSR	DOLIT
	DW	3
	JSR	TOR
GET5	JSR	DUPP
	JSR	AT
	JSR	SWAP
	JSR	CELLP
	JSR	DONXT
	DW	GET5
	JMP	DROP

;	ADCONV03 ( -- n n n n )

	DW	LINK`
	LOC
LINK`
	DB	8
	FCC	'ADCONV03'
CNV0	JSR	DOLIT
        DW      $10
CNV1	JSR	ADON
	JSR	DOLIT
	DW	$65	;ATDCTL5
	JSR	CSTOR
	JSR	WADC
	JSR	GET4
	JMP	ADOFF

;	ADCONV47 ( -- n n n n )

	DW	LINK`
	LOC
LINK`
	DB	8
	FCC	'ADCONV47'
CNV4	JSR	DOLIT
        DW      $14
	JMP	CNV1

;	READ.CURRENT ( -- n )

	DW	LINK`
	LOC
LINK`
	DB	12
	FCC	'READ.CURRENT'
RDCU    JSR     ZERO
RDCU1	JSR	CONV	
	JSR	ZERO
	JSR	DOLIT
	DW	64
	JSR	UMMOD
	JSR	SWAP
	JMP	DROP
	

;	READ.POWER ( -- n )

	DW	LINK`
	LOC
LINK`
	DB	10
	FCC	'READ.POWER'
RDPW    JSR     ONE
	JMP	RDCU1

;	READ.AC ( -- n )

	DW	LINK`
	LOC
LINK`
	DB	7
	FCC	'READ.AC'
RDAC    JSR     DOLIT
	DW	2
	JMP	RDCU1

;	READ.TUBE ( -- n )

	DW	LINK`
	LOC
LINK`
	DB	9
	FCC	'READ.TUBE'
RDTB    JSR     DOLIT
	DW	3
	JMP	RDCU1

;	READ.REGULATOR ( -- n )

	DW	LINK`
	LOC
LINK`
	DB	14
	FCC	'READ.REGULATOR'
RDRG    JSR     DOLIT
	DW	4
	JMP	RDCU1

;==============================================
;	D/A conversion

;	SPI.INIT

	DW	LINK`
	LOC
LINK`
	DB	8
	FCC	'SPI.INIT'
SPINT	MOVB	#$40,$D6	;PORTS
        MOVB    #$E2,$D7        ;DDRS
	MOVB	#$04,$D2	;SP0BR
        MOVB    #$5C,$D0        ;SP0CR1
	RTS

;	SEND.SPI ( c -- )

	DW	LINK`
	LOC
LINK`
	DB	8
	FCC	'SEND.SPI'
SNDSP	JSR	SP0DR
	JSR	CSTOR
SPIDN	JSR	DOLIT
	DW	7
	JSR	SP0SR
	JSR	QHI
	JSR	QBRAN
	DW	SPIDN
	RTS

;	SS.HI

	DW	LINK`
	LOC
LINK`
	DB	5
	FCC	'SS.HI'
SSHI	JSR	DOLIT
	DW	7
	JSR	PORTS
	JMP	HIGH

;	SS.LO

	DW	LINK`
	LOC
LINK`
	DB	5
	FCC	'SS.LO'
SSLO	JSR	DOLIT
	DW	7
	JSR	PORTS
	JMP	LOW

;       SEND.DA ( n -- )

	DW	LINK`
	LOC
LINK`
        DB      7
        FCC     'SEND.DA'
SNDDA:  JSR     DOLIT
        DW      $100
        JSR     SLMOD
        JSR     SNDSP
        JMP     SNDSP

;       DACONV ( ch# n -- )

	DW	LINK`
	LOC
LINK`
        DB      6
        FCC     'DACONV'
DACNV:  JSR     SSLO
        JSR     DOLIT
        DW      $3FF
        JSR     ANDD
        JSR     CELLS
        JSR     CELLS
        JSR     SWAP
        JSR     CELLS
        JSR     CELLS
        JSR     DOLIT
        DW      3
        JSR     ORR
        JSR     DOLIT
        DW      $1000
        JSR     STAR
        JSR     ORR
        JSR     SNDDA
        JMP     SSHI

;	SET.CURRENT ( n -- )

	DW	LINK`
	LOC
LINK`
	DB	11
        FCC     'SET.CURRENT'
CNVC	JSR	ZERO
CNVC1	JSR	SWAP
	JMP	DACNV	

;	SET.POWER ( n -- )

	DW	LINK`
	LOC
LINK`
	DB	9
	FCC	'SET.POWER'
CNVP	JSR	ONE
	JMP	CNVC1	


;       
;===============================================================
;	Laser Interface
;
;	STEPS1	( -- a )
;	  Point to motor1 step count.

	DW	LINK`
	LOC
LINK`
	DB	6
	FCC	'STEPS1'
STEPS1:
	LDD	#(RAMBASE+26)
	STD	2,-X
	RTS

;	STEPS2	( -- a )
;	  Point to motor2 step count.

	DW	LINK`
	LOC
LINK`
	DB	6
	FCC	'STEPS2'
STEPS2:
	LDD	#(RAMBASE+28)
	STD	2,-X
	RTS

;	TICKS	( -- a )
;	  Point to delay count.

	DW	LINK`
	LOC
LINK`
	DB	5
	FCC	'TICKS'
TICKS:
	LDD	#(RAMBASE+30)
	STD	2,-X
	RTS

	DW	LINK`
	LOC
LINK`
        DB      6
        FCC     'TICKS1'
TICKS1:
        LDD     #(RAMBASE+32)
	STD	2,-X
	RTS


;	INIT.SYS

	DW	LINK`
	LOC
LINK`
        DB      8
        FCC     'INIT.SYS'
INITSYS:
        JSR     SPINT
	JSR	DOLIT
	DW	$38
	JSR	DDRP
	JSR	CSTOR
	JSR	DOLIT
	DW	$33
	JSR	DDRT
        JSR     CSTOR
        JSR     DOLIT
        DW      $81
        JSR     DOLIT
        DW      $64     ;ATDCTL4
	JSR	CSTOR
	JSR	ZERO
	JSR	DUPP
	JSR	STEPS1
	JSR	STORE
	JSR	STEPS2
	JSR	STORE
	JSR	DOLIT
	DW	600
	JSR	TICKS
	JMP	STORE
	JSR	DOLIT
        DW      60
        JSR     TICKS1
	JMP	STORE
;
;	MANUAL

	DW	LINK`
	LOC
LINK`
        DB      6
        FCC     'MANUAL'
MANUAL:	JSR	DOLIT
	DW	5
PHIGH	JSR	PORTP
	JMP	HIGH
;
;	REMOTE

	DW	LINK`
	LOC
LINK`
        DB      6
        FCC     'REMOTE'
REMOTE:	JSR	DOLIT
	DW	5
PLOW	JSR	PORTP
	JMP	LOW

;	POWER.ON

	DW	LINK`
	LOC
LINK`
        DB      8
        FCC     'POWER.ON'
PWRON:	JSR	DOLIT
	DW	4
	JMP	PHIGH
;
;	POWER.OFF

	DW	LINK`
	LOC
LINK`
        DB      9
        FCC     'POWER.OFF'
PWROFF:	JSR	DOLIT
	DW	4
	JMP	PLOW

;	HV.ON

	DW	LINK`
	LOC
LINK`
        DB      5
        FCC     'HV.ON'
HVON:   JSR     DOLIT
	DW	3
	JMP	PHIGH
;
;	HV.OFF

	DW	LINK`
	LOC
LINK`
        DB      6
        FCC     'HV.OFF'
HVOFF:  JSR     DOLIT
	DW	3
	JMP	PLOW

;	MOTOR1.FORE

	DW	LINK`
	LOC
LINK`
        DB		11
        FCC     'MOTOR1.FORE'
MT1ON:	JSR	ONE
THIGH	JSR	PORTT
	JMP	HIGH

;	MOTOR1.BACK

	DW	LINK`
	LOC
LINK`
        DB      11
        FCC     'MOTOR1.BACK'
MT1OF:	JSR	ONE
TLOW	JSR	PORTT
	JMP	LOW

;	MOTOR1.STEP

	DW	LINK`
	LOC
LINK`
        DB      11
        FCC     'MOTOR1.STEP'
MT1ST:	JSR	ZERO
	JSR	PORTT
	JSR	DDUP
	JSR	HIGH
        JSR     DELAY1
	JSR	LOW
	JSR	ONE
	JSR	PORTT
	JSR	QHI
	JSR	QBRAN
	DW	M1ST2
	JSR	ONE
	BRA	M1ST3
M1ST2:	JSR	MONE
M1ST3:	JSR	STEPS1
	JSR	PSTOR
	JMP	DELAY

;	MOTOR2.FORE

	DW	LINK`
	LOC
LINK`
        DB      11
        FCC     'MOTOR2.FORE'
MT2ON:	JSR	DOLIT
	DW	5
	JMP	THIGH

;	MOTOR2.BACK

	DW	LINK`
	LOC
LINK`
        DB      11
        FCC     'MOTOR2.BACK'
MT2OF:	JSR	DOLIT
	DW	5
	JMP	TLOW

;	MOTOR2.STEP

	DW	LINK`
	LOC
LINK`
        DB      11
        FCC     'MOTOR2.STEP'
MT2ST:	JSR	DOLIT
	DW	4
	JSR	PORTT
	JSR	DDUP
	JSR	HIGH
        JSR     DELAY1
	JSR	LOW
	JSR	DOLIT
	DW	5
	JSR	PORTT
	JSR	QHI
	JSR	QBRAN
	DW	M2ST2
	JSR	ONE
	BRA	M2ST3
M2ST2:	JSR	MONE
M2ST3:	JSR	STEPS2
	JSR	PSTOR
	JMP	DELAY

;	DELAY	( -- )
;	  Do TICKS empty loops.

	DW	LINK`
	LOC
LINK`
	DB	5
	FCC	'DELAY'
DELAY:	JSR	TICKS
	JSR	AT
	JSR	TOR
DLY1:	JSR	DONXT
	DW	DLY1
	RTS

	DW	LINK`
	LOC
LINK`
        DB      6
        FCC     'DELAY1'
DELAY1:  JSR     TICKS1
	JSR	AT
	JSR	TOR
DLY2:   JSR     DONXT
        DW      DLY2
	RTS

;	MOTOR1+	( n -- )
;	  Motor1 moves n steps.

	DW	LINK`
	LOC
LINK`
	DB	7
	FCC	'MOTOR1+'
MT1P1:	JSR	MT1ON
		JSR	TOR
MT1P2:	JSR	DOLIT
		DW	3
		JSR	PORTT
		JSR	QHI
		JSR	QBRAN
		DW	MT1P3
            JSR     RFROM
		JSR	DROP
		RTS
MT1P3:  JSR     MT1ST
		JSR	DONXT
		DW	MT1P2
		RTS

;	MOTOR1-	( n -- )
;	  Motor1 moves -n steps.

	DW	LINK`
	LOC
LINK`
	DB	7
	FCC	'MOTOR1-'
MT1M1:  JSR     MT1OF
		JSR	TOR
MT1M2:	JSR	DOLIT
		DW	2
		JSR	PORTT
		JSR	QHI
		JSR	QBRAN
		DW	MT1M3
                JSR     RFROM
		JSR	DROP
		RTS
MT1M3:  JSR     MT1ST
		JSR	DONXT
		DW	MT1M2
		RTS

;	MOTOR2+	( n -- )
;	  Motor2 moves n steps.

	DW	LINK`
	LOC
LINK`
	DB	7
	FCC	'MOTOR2+'
MT2P1:	JSR	MT2ON
		JSR	TOR
MT2P2:	JSR	DOLIT
		DW	7
		JSR	PORTT
		JSR	QHI
		JSR	QBRAN
		DW	MT2P3
                JSR     RFROM
		JSR	DROP
		RTS
MT2P3:  JSR     MT2ST
		JSR	DONXT
		DW	MT2P2
		RTS

;	MOTOR2-	( n -- )
;	  Motor2 moves -n steps.

	DW	LINK`
	LOC
LINK`
	DB	7
	FCC	'MOTOR2-'
MT2M1:  JSR     MT2OF
		JSR	TOR
MT2M2:	JSR	DOLIT
		DW	6
		JSR	PORTT
		JSR	QHI
		JSR	QBRAN
		DW	MT2M3
                JSR     RFROM
		JSR	DROP
		RTS
MT2M3:  JSR     MT2ST
		JSR	DONXT
		DW	MT2M2
		RTS

;	POWER.ON?	( -- f )
;	  Return POWER.ON input.

	DW	LINK`
	LOC
LINK`
	DB	 9
	FCC	'POWER.ON?'
PONQ
	JSR	DOLIT
	DW	2
AQHI	JSR	PORTA
	JMP	QHI

;	HV.ON?	( -- f )
;	  Return HV.ON input.

	DW	LINK`
	LOC
LINK`
	DB	 6
	FCC	'HV.ON?'
HVONQ
	JSR	DOLIT
	DW	3
	JMP	AQHI

;	WATER.FLOW?	( -- f )
;	  Return WATER.FLOW input.

	DW	LINK`
	LOC
LINK`
	DB	 11
	FCC	'WATER.FLOW?'
WFQ
	JSR	DOLIT
	DW	4
	JMP	AQHI

;	WATER.TEMP?	( -- f )
;	  Return WATER.TEMP input.

	DW	LINK`
	LOC
LINK`
	DB	 11
	FCC	'WATER.TEMP?'
WTQ
	JSR	DOLIT
	DW	5
	JMP	AQHI

;	LASER.ON?	( -- f )
;	  Return LASER.ON input.

	DW	LINK`
	LOC
LINK`
	DB	 9
	FCC	'LASER.ON?'
LONQ
	JSR	DOLIT
	DW	6
	JMP	AQHI

;	COVER.OK?	( -- f )
;	  Return COVER.OK input.

	DW	LINK`
	LOC
LINK`
	DB	 9
	FCC	'COVER.OK?'
COKQ
	JSR	DOLIT
	DW	7
	JMP	AQHI

;	STANDBY?	( -- f )
;	  Return STANDBY input.

	DW	LINK`
	LOC
LINK`
	DB	 8
	FCC	'STANDBY?'
STBY
	JSR	ZERO
	JSR	PORTB
	JMP	QHI

;	LINE1?	( -- f )
;	  Return LINE1 input.

	DW	LINK`
	LOC
LINK`
	DB	 6
	FCC	'LINE1?'
L1Q
	JSR	DOLIT
	DW	5
ADQHI	JSR	PORTAD
	JMP	QHI

;	LINE2?	( -- f )
;	  Return LINE2 input.

	DW	LINK`
	LOC
LINK`
	DB	 6
	FCC	'LINE2?'
L2Q
	JSR	DOLIT
	DW	6
	JMP	ADQHI

;	LINE3?	( -- f )
;	  Return LINE3 input.

	DW	LINK`
	LOC
LINK`
	DB	 6
	FCC	'LINE3?'
L3Q
	JSR	DOLIT
	DW	7
	JMP	ADQHI



;       
;===============================================================

LASTN   EQU	LINK`   ;last name defined
	END	


